// Utilities
//-----------
var arrayToString = function(a, before, after) {
  var ret = "";
  for (var i in a) {
    ret += (before || "") + "#" + i + " " + a[i].toString() + (after || "");
  }
  return ret;
};

var objToString = function(o, before, after) {
  var ret = "";
  for (var k in o) {
    ret += (before || "") + k + " => " + o[k].toString() + (after || "");
  }
  return ret;
};

// Universe
//----------
var Universe = function() {
  this.dict = {};
};

Universe.getDictKey_ = function(type, key) {
  return type + "-" + key;
};

Universe.prototype.put = function(type, key, value) {
  this.dict[Universe.getDictKey_(type, key)] = value;
};

Universe.prototype.get = function(type, key) {
  return this.dict[Universe.getDictKey_(type, key)];
};

Universe.prototype.produce = function() {
  var products = [];
  for (var k in this.dict) {
    var b = this.dict[k];
    var bproducts = b.produce();
    if (bproducts) {
      for (var i in bproducts) {
        products.push(bproducts[i]);
      }
    }
  }
  return products;
};

Universe.prototype.transfer = function(products) {
  var remains = [];
  var consumes = [];
  for (var pi in products) {
    var product = products[pi];
    var transfers = product.src.getTransfers(product.elem);
    for (var t in transfers) {
      var transfer = transfers[t];
      var transferQty = Math.max(product.qty, transfer.qpr);
      if (transferQty > 0) {
        consumes.push({ dest: transfer.dest, elem: product.elem, qty: transferQty});
      }
      product.qty -= transferQty;
    }
    if (product.qty > 0) {
      remains.push(product);
    }
  }
  return [remains, consumes];
};

Universe.prototype.toString = function() {
  return "[Universe \n" + objToString(this.dict, "  ", "\n") + "]";
};

var U = new Universe();


// A 'living' game object
//------------------------
var Being = function() {};

Being.prototype._init_ = function(type, key, umass, temp, cc, qty) {
  this.type = type;
  this.key = key;
  this.umass = umass;  // The unitary mass of the thing (in kg)
  this.temp = temp;    // Its temperature (in deg)
  this.cc = cc;        // Its calorific capacity in deg/kg/round
  this.qty = qty || 1; // The number of things, 1 by default
  this.prods = [];
  this.transfers = [];
  U.put(type, key, this);
};

/* Probably not needed. Probably... 
Being.create = function(type, key, umass, temp, cc, qty) {
  var ret = new Being();
  ret._init_(type, key, umass, temp, cc, qty);
  return ret;
};
*/

Being.prototype.addProduction = function(prod) {
  this.prods.push(prod);
};

Being.prototype.addTransfer = function(elem, dest, qpr) {
  var transfer;
  for (var i in this.transfers) {
    transfer = this.transfers[i];
    if ((transfer.elem === elem) && (transfer.dest === dest)) {
      transfer.qpr += qpr;
      if (transfer.qpr === 0) {
        this.transfers.splice(i, 1);
      }
      return;
    }
  }
  transfer = new Transfer(elem, this, dest, qpr);
  this.transfers.push(transfer);
};

Being.prototype.hasProductionLeft = function(elem) {
  for (var i in this.prods) {
    var production = this.prods[i];
    var transfers = this.getTransfers(production.elem);
    var sum = 0;
    for (var t in transfers) {
      sum += transfers[t].qpr;
    }
    if (sum < production.qpr) {
      return true;
    }
  }
  return false;
};

Being.prototype.getTransfers = function(elem) {
  var ret = [];
  for (var i in this.transfers) {
    var transfer = this.transfers[i];
    if (transfer.elem === elem) {
      ret.push(transfer);
    }
  }
  return ret;
};

Being.prototype.produce = function() {
  var products = [];
  for (var i in this.prods) {
    var production = this.prods[i];
    var product = production.produce();
    if (product) {
      products.push(product);
    }
  }
  return products;
};

Being.prototype.transferHeat = function(heat) {
  this.temp += heat / (this.umass * this.cc);
};

Being.prototype.isAlright = function() {
  return true;
};

Being.prototype.toString = function() {
  return "[Being type=" + this.type + ", key=" + this.key + ", mass=" + this.umass +
    ", temp=" + this.temp + ", cc=" + this.cc + ", qty=" + this.qty +
    "\n prod=[\n" + arrayToString(this.prods, "  ", "\n") + " ]" +
    "\n trans=[\n" + arrayToString(this.transfers, "  ", "\n") + " ]\n]";
};


// A world
//---------
var World = function() {};
World.prototype = new Being();
World.TYPE = "World";

World.prototype._init_ = function(name, mass, temp, cc, freeze, boil) {
  Being.prototype._init_.call(this, World.TYPE, name, mass, temp, cc);
  this.name = name;
  this.freeze = freeze;
  this.boil = boil;
};

World.create = function(name, mass, temp, cc, freeze, boil) {
  var ret = new World();
  ret._init_(name, mass, temp, cc, freeze, boil);
  return ret;
};

World.prototype.freezes = function() {
  return this.temp < this.freeze;
};

World.prototype.boils = function() {
  return this.temp > this.boil;
};

World.prototype.isAlright = function() {
  return !(this.freezes() || this.boils());
};

World.prototype.consume = function(elem, qty) {
  //console.log(this.name + " consuming " + qty + "  " + elem.name);
  if (elem === Elem.FIRE) {
    this.transferHeat(qty * 250000);
  }
};

World.prototype.toString = function() {
  var ret = "[World" + Being.prototype.toString.call(this) + "]";
  return ret;
};

// A produced and transferred element
//------------------------------------
var Elem = function() {};

Elem.prototype._init_ = function(name) {
  this.name = name;
};

Elem.prototype.toString = function() {
  return "[Element name=" + this.name + "]";
};

// Actual elements
//-----------------
var Fire = function() {};
Fire.prototype = new Elem();

Fire.prototype._init_ = function() {
  Elem.prototype._init_.call(this, "Fire");
};

Fire.create = function() {
  var ret = new Fire();
  ret._init_();
  return ret;
};

Elem.FIRE = Fire.create();

var Water = function() {};
Water.prototype = new Elem();

Water.prototype._init_ = function() {
  Elem.prototype._init_.call(this, "Water");
};

Water.create = function() {
  var ret = new Water();
  ret._init_();
  return ret;
};

Elem.WATER = Water.create();

// A generic production
//----------------------
var Production = function() {};
Production.prototype._init_ = function(src, elem, qpr) {
  this.src = src;   // Producer thing
  this.elem = elem; // Element produced
  this.qpr = qpr;   // Quantity per round
};

Production.create = function(src, elem, qpr) {
  var ret = new Production();
  ret._init_(src, elem, qpr);
  return ret;
};

Production.prototype.produce = function() {
  return new Product(this.src, this.elem, this.qpr);
};

Production.prototype.consume = function(dest) {
  //console.log("consume called on " + this.toString() + " to destination " + dest.toString());
};

Production.prototype.toString = function() {
  return "[Production src=" + this.src.name + ", elem=" + this.elem.name + ", qpr=" + this.qpr + "]";
};

// Fire production: removes heat from the source
var FireProduction = function() {};
FireProduction.prototype = new Production();

FireProduction.prototype._init_ = function(src, qpr) {
  Production.prototype._init_.call(this, src, Elem.FIRE, qpr);
};

FireProduction.create = function(src, qpr) {
  var ret = new FireProduction();
  ret._init_(src, qpr);
  return ret;
};

FireProduction.prototype.produce = function() {
  if (!this.src.freezes()) {
    var product = Production.prototype.produce.call(this);
    this.src.transferHeat(-this.qpr * 250000);
    return product;
  }
};

// Water production
var WaterProduction = function() {};
WaterProduction.prototype = new Production();

WaterProduction.prototype._init_ = function(src, qpr) {
  Production.prototype._init_.call(this, src, Elem.WATER, qpr);
};

WaterProduction.create = function(src, qpr) {
  var ret = new WaterProduction();
  ret._init_(src, qpr);
  return ret;
};

WaterProduction.prototype.produce = function() {
  var product = Production.prototype.produce.call(this);
  //this.src.removeHeat(this.qpr);
  return product;
};


// A product (what is produced in a round by a Production)
var Product = function(src, elem, qty) {
  this.src = src;
  this.elem = elem;
  this.qty = qty;
};

Product.prototype.toString = function() {
  return "[Product src=" + this.src.name + ", elem=" + this.elem.name + ", qty=" + this.qty + "]";
};

// A user-decided transfer between worlds
var Transfer = function(elem, src, dest, qpr) {
  this.elem = elem; // Transferred element
  this.src = src;   // Source world
  this.dest = dest; // Destination world
  this.qpr = qpr;   // Quantity per round
};

Transfer.prototype.toString = function() {
  return "[Transfer " + this.qpr + " " + this.elem.name + " from " + this.src.name +
    " to " + this.dest.name + "]";
};

// Actual worlds
//---------------
var WaterWorld = World.create("WaterWorld", 5000000, 100, 3, 250, 350);
var FireWorld = World.create("FireWorld", 10000000, 7000, 10, 5000, 10000);
var EarthWorld = World.create("EarthWorld", 7000000, 120, 5, 250, 350);
var AirWorld = World.create("AirWorld", 1000000, 0, 1, 100, 500);

FireWorld.addProduction(FireProduction.create(FireWorld, 12));
WaterWorld.addProduction(WaterProduction.create(WaterWorld, 10));

// Logic happening each round
//----------------------------
Universe.prototype.doRound = function() {
  // Produce
  var products = this.produce();
  
  // Transfer
  var remAndCons = this.transfer(products);
  var remains = remAndCons[0];
  var consumes = remAndCons[1];

  // Give back remains
  for (var r in remains) {
    var remain = remains[r];
    remain.src.consume(remain.elem, remain.qty);
  }
  
  // Consume
  for (var c in consumes) {
    var consume = consumes[c];
    consume.dest.consume(consume.elem, consume.qty);
  }
  //console.log("Fire: " + FireWorld.temp + ", Water: " + WaterWorld.temp);
};

Universe.prototype.isAlright = function() {
  for (var k in this.dict) {
    var b = this.dict[k];
    if (!b.isAlright()) {
      return false;
    }
  }
  return true;
};
